@@ -1,4 +1,23 @@
 #!/usr/bin/env python
+# -*- coding: utf-8 -*-
+#
+# Copyright (C) 2004-2013 Edgewall Software
+# Copyright (C) 2004 Dmitry Yusupov <dmitry_yus@yahoo.com>
+# Copyright (C) 2004 Mark Rowe <mrowe@bluewire.net.nz>
+# Copyright (C) 2005 Bill Soudan <bill@soudan.net>
+# Copyright (C) 2005 Florent Guillaume <fg@nuxeo.com>
+# Copyright (C) 2005 Jeroen Ruigrok van der Werven <asmodai@in-nomine.org>
+# Copyright (C) 2010 Jeff Moreland <hou5e@hotmail.com>
+#
+# All rights reserved.
+#
+# This software is licensed as described in the file COPYING, which
+# you should have received as part of this distribution. The terms
+# are also available at http://trac.edgewall.com/license.html.
+#
+# This software consists of voluntary contributions made by many
+# individuals. For the exact contribution history, see the revision
+# history and logs, available at http://trac.edgewall.org/.
 
 """
 Import a Bugzilla items into a Trac database.
@@ -6,19 +25,14 @@ Import a Bugzilla items into a Trac database.
 Requires:  Trac 0.9b1 from http://trac.edgewall.org/
            Python 2.3 from http://www.python.org/
            MySQL >= 3.23 from http://www.mysql.org/
+           or PostGreSQL 8.4 from http://www.postgresql.org/
+           or SQLite 3 from http://www.sqlite.org/
 
-Thanks:    Mark Rowe <mrowe@bluewire.net.nz>
-            for original TracDatabase class
-
-Copyright 2004, Dmitry Yusupov <dmitry_yus@yahoo.com>
-
-Many enhancements, Bill Soudan <bill@soudan.net>
-Other enhancements, Florent Guillaume <fg@nuxeo.com>
-Reworked, Jeroen Ruigrok van der Werven <asmodai@tendra.org>
-
-$Id$
+$Id: bugzilla2trac.py 12500 2014-02-12 20:54:59Z rjollos $
 """
 
+from __future__ import with_statement
+
 import re
 
 ###
@@ -28,12 +42,13 @@ import re
 # Bugzilla version.  You can find this in Bugzilla's globals.pl file.
 #
 # Currently, the following bugzilla versions are known to work:
-#   2.11 (2110), 2.16.5 (2165), 2.18.3 (2183), 2.19.1 (2191)
+#   2.11 (2110), 2.16.5 (2165), 2.16.7 (2167),  2.18.3 (2183), 2.19.1 (2191),
+#   2.23.3 (2233), 3.04.4 (3044)
 #
 # If you run this script on a version not listed here and it is successful,
-# please report it to the Trac mailing list and drop a note to
-# asmodai@tendra.org so we can update the list.
-BZ_VERSION = 2180
+# please file a ticket at http://trac.edgewall.org
+#
+BZ_VERSION = 3044
 
 # MySQL connection parameters for the Bugzilla database.  These can also
 # be specified on the command line.
@@ -99,6 +114,10 @@ IGNORE_PRODUCTS = []
 # These milestones are ignored
 IGNORE_MILESTONES = ["---"]
 
+# Don't import user names and passwords into htpassword if
+# user is disabled in bugzilla? (i.e. profiles.DisabledText<>'')
+IGNORE_DISABLED_USERS = True
+
 # These logins are converted to these user ids
 LOGIN_MAP = {
     #'some.user@example.com': 'someuser',
@@ -137,6 +156,13 @@ KEYWORDS_MAPPING = {
 # even if not mentionned in KEYWORDS_MAPPING.
 MAP_ALL_KEYWORDS = True
 
+# Custom field mappings
+CUSTOMFIELD_MAP = {
+    #'Bugzilla_field_name': 'Trac_customfield_name',
+    #'op_sys': 'os',
+    #'cf_featurewantedby': 'wanted_by',
+    #'product': 'product'
+}
 
 # Bug comments that should not be imported.  Each entry in list should
 # be a regular expression.
@@ -179,7 +205,9 @@ STATUS_KEYWORDS = {
 IGNORED_ACTIVITY_FIELDS = ["everconfirmed"]
 
 # Regular expression and its replacement
-BUG_NO_RE = re.compile(r"\b(bug #?)([0-9])")
+# this expression will update references to bugs 1 - 99999 that
+# have the form "bug 1" or "bug #1"
+BUG_NO_RE = re.compile(r"\b(bug #?)([0-9]{1,5})\b", re.I)
 BUG_NO_REPL = r"#\2"
 
 ###
@@ -193,11 +221,8 @@ import StringIO
 
 import MySQLdb
 import MySQLdb.cursors
-try:
-    from trac.env import Environment
-except:
-    from trac.Environment import Environment
 from trac.attachment import Attachment
+from trac.env import Environment
 
 if not hasattr(sys, 'setdefaultencoding'):
     reload(sys)
@@ -214,7 +239,7 @@ sys.setdefaultencoding('latin1')
 # mapping, just return string, otherwise return value
 class FieldTranslator(dict):
     def __getitem__(self, item):
-        if not dict.has_key(self, item):
+        if item not in self:
             return item
 
         return dict.__getitem__(self, item)
@@ -224,18 +249,14 @@ statusXlator = FieldTranslator(STATUS_TRANSLATE)
 class TracDatabase(object):
     def __init__(self, path):
         self.env = Environment(path)
-        self._db = self.env.get_db_cnx()
-        self._db.autocommit = False
         self.loginNameCache = {}
         self.fieldNameCache = {}
-
-    def db(self):
-        return self._db
+        from trac.db.api import DatabaseManager
+        self.using_postgres = \
+            DatabaseManager(self.env).connection_uri.startswith("postgres:")
 
     def hasTickets(self):
-        c = self.db().cursor()
-        c.execute("SELECT count(*) FROM Ticket")
-        return int(c.fetchall()[0][0]) > 0
+        return int(self.env.db_query("SELECT count(*) FROM ticket")[0][0] > 0)
 
     def assertNoTickets(self):
         if self.hasTickets():
@@ -245,167 +266,180 @@ class TracDatabase(object):
         """Remove all severities, set them to `s`"""
         self.assertNoTickets()
 
-        c = self.db().cursor()
-        c.execute("DELETE FROM enum WHERE type='severity'")
-        for value, i in s:
-            print "  inserting severity '%s' - '%s'" % (value, i)
-            c.execute("""INSERT INTO enum (type, name, value)
-                                   VALUES (%s, %s, %s)""",
-                      ("severity", value.encode('utf-8'), i))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("DELETE FROM enum WHERE type='severity'")
+            for value, i in s:
+                print "  inserting severity '%s' - '%s'" % (value, i)
+                db("""INSERT INTO enum (type, name, value)
+                      VALUES (%s, %s, %s)""",
+                   ("severity", value, i))
 
     def setPriorityList(self, s):
         """Remove all priorities, set them to `s`"""
         self.assertNoTickets()
 
-        c = self.db().cursor()
-        c.execute("DELETE FROM enum WHERE type='priority'")
-        for value, i in s:
-            print "  inserting priority '%s' - '%s'" % (value, i)
-            c.execute("""INSERT INTO enum (type, name, value)
-                                   VALUES (%s, %s, %s)""",
-                      ("priority", value.encode('utf-8'), i))
-        self.db().commit()
-
+        with self.env.db_transaction as db:
+            db("DELETE FROM enum WHERE type='priority'")
+            for value, i in s:
+                print "  inserting priority '%s' - '%s'" % (value, i)
+                db("INSERT INTO enum (type, name, value) VALUES (%s, %s, %s)",
+                   ("priority", value, i))
 
     def setComponentList(self, l, key):
         """Remove all components, set them to `l`"""
         self.assertNoTickets()
 
-        c = self.db().cursor()
-        c.execute("DELETE FROM component")
-        for comp in l:
-            print "  inserting component '%s', owner '%s'" % \
-                            (comp[key], comp['owner'])
-            c.execute("INSERT INTO component (name, owner) VALUES (%s, %s)",
-                      (comp[key].encode('utf-8'),
-                       comp['owner'].encode('utf-8')))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("DELETE FROM component")
+            for comp in l:
+                print "  inserting component '%s', owner '%s'" % \
+                                (comp[key], comp['owner'])
+                db("INSERT INTO component (name, owner) VALUES (%s, %s)",
+                   (comp[key], comp['owner']))
 
     def setVersionList(self, v, key):
         """Remove all versions, set them to `v`"""
         self.assertNoTickets()
 
-        c = self.db().cursor()
-        c.execute("DELETE FROM version")
-        for vers in v:
-            print "  inserting version '%s'" % (vers[key])
-            c.execute("INSERT INTO version (name) VALUES (%s)",
-                      (vers[key].encode('utf-8'),))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("DELETE FROM version")
+            for vers in v:
+                print "  inserting version '%s'" % (vers[key])
+                db("INSERT INTO version (name) VALUES (%s)",
+                   (vers[key],))
 
     def setMilestoneList(self, m, key):
         """Remove all milestones, set them to `m`"""
         self.assertNoTickets()
 
-        c = self.db().cursor()
-        c.execute("DELETE FROM milestone")
-        for ms in m:
-            milestone = ms[key]
-            print "  inserting milestone '%s'" % (milestone)
-            c.execute("INSERT INTO milestone (name) VALUES (%s)",
-                      (milestone.encode('utf-8'),))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("DELETE FROM milestone")
+            for ms in m:
+                milestone = ms[key]
+                print "  inserting milestone '%s'" % (milestone)
+                db("INSERT INTO milestone (name) VALUES (%s)",
+                   (milestone,))
 
     def addTicket(self, id, time, changetime, component, severity, priority,
                   owner, reporter, cc, version, milestone, status, resolution,
-                  summary, description, keywords):
-        c = self.db().cursor()
+                  summary, description, keywords, customfields):
 
-        desc = description.encode('utf-8')
+        desc = description
         type = "defect"
 
-        if severity.lower() == "enhancement":
+        if SEVERITIES:
+            if severity.lower() == "enhancement":
                 severity = "minor"
                 type = "enhancement"
 
+        else:
+            if priority.lower() == "enhancement":
+                priority = "minor"
+                type = "enhancement"
+
         if PREFORMAT_COMMENTS:
-          desc = '{{{\n%s\n}}}' % desc
+            desc = '{{{\n%s\n}}}' % desc
 
         if REPLACE_BUG_NO:
             if BUG_NO_RE.search(desc):
                 desc = re.sub(BUG_NO_RE, BUG_NO_REPL, desc)
 
-        if PRIORITIES_MAP.has_key(priority):
+        if priority in PRIORITIES_MAP:
             priority = PRIORITIES_MAP[priority]
 
         print "  inserting ticket %s -- %s" % (id, summary)
 
-        c.execute("""INSERT INTO ticket (id, type, time, changetime, component,
-                                         severity, priority, owner, reporter,
-                                         cc, version, milestone, status,
-                                         resolution, summary, description,
-                                         keywords)
-                                 VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s,
-                                         %s, %s, %s, %s, %s, %s, %s, %s)""",
-                  (id, type.encode('utf-8'), time.strftime('%s'),
-                   changetime.strftime('%s'), component.encode('utf-8'),
-                   severity.encode('utf-8'), priority.encode('utf-8'), owner,
-                   reporter, cc, version, milestone.encode('utf-8'),
-                   status.lower(), resolution, summary.encode('utf-8'), desc,
-                   keywords))
+        with self.env.db_transaction as db:
+            db("""INSERT INTO ticket (id, type, time, changetime, component,
+                                      severity, priority, owner, reporter, cc,
+                                      version, milestone, status, resolution,
+                                      summary, description, keywords)
+                  VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
+                          %s, %s, %s, %s)
+                  """, (id, type, datetime2epoch(time),
+                        datetime2epoch(changetime), component, severity,
+                        priority, owner, reporter, cc, version, milestone,
+                        status.lower(), resolution, summary, desc, keywords))
 
-        self.db().commit()
-        return self.db().get_last_id(c, 'ticket')
+        if self.using_postgres:
+            with self.env.db_transaction as db:
+                c = db.cursor()
+                c.execute("""
+                    SELECT SETVAL('ticket_id_seq', MAX(id)) FROM ticket;
+                    SELECT SETVAL('report_id_seq', MAX(id)) FROM report""")
+                ticket_id = db.get_last_id(c, 'ticket')
+
+        # add all custom fields to ticket
+        for name, value in customfields.iteritems():
+            self.addTicketCustomField(ticket_id, name, value)
+
+        return ticket_id
+
+    def addTicketCustomField(self, ticket_id, field_name, field_value):
+        if field_value == None:
+            return
+        self.env.db_transaction("""
+            INSERT INTO ticket_custom (ticket, name, value) VALUES (%s, %s, %s)
+            """, (ticket_id, field_name, field_value))
 
     def addTicketComment(self, ticket, time, author, value):
-        comment = value.encode('utf-8')
+        comment = value
 
         if PREFORMAT_COMMENTS:
-          comment = '{{{\n%s\n}}}' % comment
+            comment = '{{{\n%s\n}}}' % comment
 
         if REPLACE_BUG_NO:
             if BUG_NO_RE.search(comment):
                 comment = re.sub(BUG_NO_RE, BUG_NO_REPL, comment)
 
-        c = self.db().cursor()
-        c.execute("""INSERT INTO ticket_change (ticket, time, author, field,
-                                                oldvalue, newvalue)
-                                        VALUES (%s, %s, %s, %s, %s, %s)""",
-                  (ticket, time.strftime('%s'), author, 'comment', '', comment))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("""INSERT INTO ticket_change (ticket, time, author, field,
+                                             oldvalue, newvalue)
+                  VALUES (%s, %s, %s, %s, %s, %s)
+                  """, (ticket, datetime2epoch(time), author, 'comment', '',
+                        comment))
 
     def addTicketChange(self, ticket, time, author, field, oldvalue, newvalue):
-        c = self.db().cursor()
 
         if field == "owner":
-            if LOGIN_MAP.has_key(oldvalue):
+            if oldvalue in LOGIN_MAP:
                 oldvalue = LOGIN_MAP[oldvalue]
-            if LOGIN_MAP.has_key(newvalue):
+            if newvalue in LOGIN_MAP:
                 newvalue = LOGIN_MAP[newvalue]
 
         if field == "priority":
-            if PRIORITIES_MAP.has_key(oldvalue.lower()):
+            if oldvalue.lower() in PRIORITIES_MAP:
                 oldvalue = PRIORITIES_MAP[oldvalue.lower()]
-            if PRIORITIES_MAP.has_key(newvalue.lower()):
+            if newvalue.lower() in PRIORITIES_MAP:
                 newvalue = PRIORITIES_MAP[newvalue.lower()]
 
         # Doesn't make sense if we go from highest -> highest, for example.
         if oldvalue == newvalue:
             return
 
-        c.execute("""INSERT INTO ticket_change (ticket, time, author, field,
-                                                oldvalue, newvalue)
-                                        VALUES (%s, %s, %s, %s, %s, %s)""",
-                  (ticket, time.strftime('%s'), author, field,
-                   oldvalue.encode('utf-8'), newvalue.encode('utf-8')))
-        self.db().commit()
+        with self.env.db_transaction as db:
+            db("""INSERT INTO ticket_change (ticket, time, author, field,
+                                             oldvalue, newvalue)
+                  VALUES (%s, %s, %s, %s, %s, %s)
+                  """, (ticket, datetime2epoch(time), author, field,
+                        oldvalue, newvalue))
 
     def addAttachment(self, author, a):
-        description = a['description'].encode('utf-8')
-        id = a['bug_id']
-        filename = a['filename'].encode('utf-8')
-        filedata = StringIO.StringIO(a['thedata'])
-        filesize = len(filedata.getvalue())
-        time = a['creation_ts']
-        print "    ->inserting attachment '%s' for ticket %s -- %s" % \
-                (filename, id, description)
-
-        attachment = Attachment(self.env, 'ticket', id)
-        attachment.author = author
-        attachment.description = description
-        attachment.insert(filename, filedata, filesize, time.strftime('%s'))
-        del attachment
+        if a['filename'] != '':
+            description = a['description']
+            id = a['bug_id']
+            filename = a['filename']
+            filedata = StringIO.StringIO(a['thedata'])
+            filesize = len(filedata.getvalue())
+            time = a['creation_ts']
+            print "    ->inserting attachment '%s' for ticket %s -- %s" % \
+                    (filename, id, description)
+            attachment = Attachment(self.env, 'ticket', id)
+            attachment.author = author
+            attachment.description = description
+            attachment.insert(filename, filedata, filesize,
+                              datetime2epoch(time))
+            del attachment
 
     def getLoginName(self, cursor, userid):
         if userid not in self.loginNameCache:
@@ -427,8 +461,14 @@ class TracDatabase(object):
 
     def getFieldName(self, cursor, fieldid):
         if fieldid not in self.fieldNameCache:
-            cursor.execute("SELECT * FROM fielddefs WHERE fieldid = %s",
-                           (fieldid))
+            # fielddefs.fieldid got changed to fielddefs.id in Bugzilla
+            # 2.23.3.
+            if BZ_VERSION >= 2233:
+                cursor.execute("SELECT * FROM fielddefs WHERE id = %s",
+                               (fieldid))
+            else:
+                cursor.execute("SELECT * FROM fielddefs WHERE fieldid = %s",
+                               (fieldid))
             fieldName = cursor.fetchall()
 
             if fieldName:
@@ -449,8 +489,9 @@ def makeWhereClause(fieldName, values, negative=False):
         connector, op = ' AND ', '!='
     else:
         connector, op = ' OR ', '='
-    clause = connector.join(["%s %s '%s'" % (fieldName, op, value) for value in values])
-    return ' ' + clause
+    clause = connector.join(["%s %s '%s'" % (fieldName, op, value)
+                             for value in values])
+    return ' (' + clause + ')'
 
 def convert(_db, _host, _user, _password, _env, _force):
     activityFields = FieldTranslator()
@@ -466,7 +507,8 @@ def convert(_db, _host, _user, _password, _env, _force):
             (_db, _host, _user, ("*" * len(_password)))
     mysql_con = MySQLdb.connect(host=_host,
                 user=_user, passwd=_password, db=_db, compress=1,
-                cursorclass=MySQLdb.cursors.DictCursor)
+                cursorclass=MySQLdb.cursors.DictCursor,
+                charset='utf8')
     mysql_cur = mysql_con.cursor()
 
     # init Trac environment
@@ -476,16 +518,13 @@ def convert(_db, _host, _user, _password, _env, _force):
     # force mode...
     if _force == 1:
         print "\nCleaning all tickets..."
-        c = trac.db().cursor()
-        c.execute("DELETE FROM ticket_change")
-        trac.db().commit()
-
-        c.execute("DELETE FROM ticket")
-        trac.db().commit()
-
-        c.execute("DELETE FROM attachment")
-	attachments_dir = os.path.join(os.path.normpath(trac.env.path),
-                                "attachments")
+        with trac.env.db_transaction as db:
+            db("DELETE FROM ticket_change")
+            db("DELETE FROM ticket")
+            db("DELETE FROM ticket_custom")
+            db("DELETE FROM attachment")
+        attachments_dir = os.path.join(os.path.normpath(trac.env.path),
+                                       "attachments")
         # Straight from the Python documentation.
         for root, dirs, files in os.walk(attachments_dir, topdown=False):
             for name in files:
@@ -494,7 +533,6 @@ def convert(_db, _host, _user, _password, _env, _force):
                 os.rmdir(os.path.join(root, name))
         if not os.stat(attachments_dir):
             os.mkdir(attachments_dir)
-        trac.db().commit()
         print "All tickets cleaned..."
 
 
@@ -535,13 +573,16 @@ def convert(_db, _host, _user, _password, _env, _force):
         trac.setComponentList(components, 'name')
     else:
         if BZ_VERSION >= 2180:
-            sql = """SELECT p.name AS product, c.name AS comp, c.initialowner AS owner
-                   FROM components c, products p"""
-            sql += " WHERE c.product_id = p.id and " + makeWhereClause('p.name', PRODUCTS)
+            sql = ("SELECT p.name AS product, c.name AS comp, "
+                   " c.initialowner AS owner "
+                   "FROM components c, products p "
+                   "WHERE c.product_id = p.id AND" +
+                   makeWhereClause('p.name', PRODUCTS))
         else:
-            sql = """SELECT program AS product, value AS comp, initialowner AS owner
-                   FROM components"""
-            sql += " WHERE" + makeWhereClause('program', PRODUCTS)
+            sql = ("SELECT program AS product, value AS comp, "
+                   " initialowner AS owner "
+                   "FROM components WHERE" +
+                   makeWhereClause('program', PRODUCTS))
         mysql_cur.execute(sql)
         lines = mysql_cur.fetchall()
         all_components = {} # product -> components
@@ -572,9 +613,10 @@ def convert(_db, _host, _user, _password, _env, _force):
 
     print "\n4. Import versions..."
     if BZ_VERSION >= 2180:
-        sql = """SELECT DISTINCTROW versions.value AS value
-                               FROM products, versions"""
-        sql += " WHERE" + makeWhereClause('products.name', PRODUCTS)
+        sql = """SELECT DISTINCTROW v.value AS value
+                               FROM products p, versions v"""
+        sql += " WHERE v.product_id = p.id AND"
+        sql += makeWhereClause('p.name', PRODUCTS)
     else:
         sql = "SELECT DISTINCTROW value FROM versions"
         sql += " WHERE" + makeWhereClause('program', PRODUCTS)
@@ -590,18 +632,27 @@ def convert(_db, _host, _user, _password, _env, _force):
     trac.setMilestoneList(milestones, 'value')
 
     print "\n6. Retrieving bugs..."
-    sql = """SELECT DISTINCT b.*, c.name AS component, p.name AS product
-                        FROM bugs AS b, components AS c, products AS p """
-    sql += " WHERE (" + makeWhereClause('p.name', PRODUCTS)
-    sql += ") AND b.product_id = p.id"
-    sql += " AND b.component_id = c.id"
-    sql += " ORDER BY b.bug_id"
+    if BZ_VERSION >= 2180:
+        sql = """SELECT DISTINCT b.*, c.name AS component, p.name AS product
+                            FROM bugs AS b, components AS c, products AS p """
+        sql += " WHERE" + makeWhereClause('p.name', PRODUCTS)
+        sql += " AND b.product_id = p.id"
+        sql += " AND b.component_id = c.id"
+        sql += " ORDER BY b.bug_id"
+    else:
+        sql = """SELECT DISTINCT b.*, c.value AS component, p.product AS product
+                            FROM bugs AS b, components AS c, products AS p """
+        sql += " WHERE" + makeWhereClause('p.product', PRODUCTS)
+        sql += " AND b.product = p.product"
+        sql += " AND b.component = c.value"
+        sql += " ORDER BY b.bug_id"
     mysql_cur.execute(sql)
     bugs = mysql_cur.fetchall()
 
 
     print "\n7. Import bugs and bug activity..."
     for bug in bugs:
+
         bugid = bug['bug_id']
 
         ticket = {}
@@ -613,18 +664,32 @@ def convert(_db, _host, _user, _password, _env, _force):
             ticket['component'] = bug['product']
         else:
             ticket['component'] = bug['component']
-        ticket['severity'] = bug['bug_severity']
-        ticket['priority'] = bug['priority'].lower()
+
+        if SEVERITIES:
+            ticket['severity'] = bug['bug_severity']
+            ticket['priority'] = bug['priority'].lower()
+        else:
+            # use bugzilla severities as trac priorities, and ignore bugzilla
+            # priorities
+            ticket['severity'] = ''
+            ticket['priority'] = bug['bug_severity']
 
         ticket['owner'] = trac.getLoginName(mysql_cur, bug['assigned_to'])
         ticket['reporter'] = trac.getLoginName(mysql_cur, bug['reporter'])
 
+        # pack bugzilla fields into dictionary of trac custom field
+        # names and values
+        customfields = {}
+        for bugfield, customfield in CUSTOMFIELD_MAP.iteritems():
+            customfields[customfield] = bug[bugfield]
+        ticket['customfields'] = customfields
+
         mysql_cur.execute("SELECT * FROM cc WHERE bug_id = %s", bugid)
         cc_records = mysql_cur.fetchall()
         cc_list = []
         for cc in cc_records:
             cc_list.append(trac.getLoginName(mysql_cur, cc['who']))
-        cc_list = [cc for cc in cc_list if '@' in cc and cc not in IGNORE_CC]
+        cc_list = [cc for cc in cc_list if cc not in IGNORE_CC]
         ticket['cc'] = string.join(cc_list, ', ')
 
         ticket['version'] = bug['version']
@@ -664,7 +729,7 @@ def convert(_db, _host, _user, _password, _env, _force):
                     ignore = True
 
             if ignore:
-                    continue
+                continue
 
             trac.addTicketComment(ticket=bugid,
                 time = desc['bug_when'],
@@ -672,7 +737,7 @@ def convert(_db, _host, _user, _password, _env, _force):
                 value = desc['thetext'])
 
         mysql_cur.execute("""SELECT * FROM bugs_activity WHERE bug_id = %s
-                           ORDER BY bug_when""" % bugid)
+                             ORDER BY bug_when""" % bugid)
         bugs_activity = mysql_cur.fetchall()
         resolution = ''
         ticketChanges = []
@@ -697,7 +762,10 @@ def convert(_db, _host, _user, _password, _env, _force):
 
             # convert bugzilla field names...
             if field_name == "bug_severity":
-                field_name = "severity"
+                if SEVERITIES:
+                    field_name = "severity"
+                else:
+                    field_name = "priority"
             elif field_name == "assigned_to":
                 field_name = "owner"
             elif field_name == "bug_status":
@@ -765,18 +833,19 @@ def convert(_db, _host, _user, _password, _env, _force):
 
             # Bugzilla splits large summary changes into two records.
             for oldChange in ticketChanges:
-              if (field_name == "summary"
-                  and oldChange['field'] == ticketChange['field']
-                  and oldChange['time'] == ticketChange['time']
-                  and oldChange['author'] == ticketChange['author']):
-                  oldChange['oldvalue'] += " " + ticketChange['oldvalue']
-                  oldChange['newvalue'] += " " + ticketChange['newvalue']
-                  break
-              # cc sometime appear in different activities with same time
-              if (field_name == "cc" \
-                  and oldChange['time'] == ticketChange['time']):
-                  oldChange['newvalue'] += ", " + ticketChange['newvalue']
-                  break
+                if (field_name == "summary"
+                    and oldChange['field'] == ticketChange['field']
+                    and oldChange['time'] == ticketChange['time']
+                    and oldChange['author'] == ticketChange['author']):
+                    oldChange['oldvalue'] += " " + ticketChange['oldvalue']
+                    oldChange['newvalue'] += " " + ticketChange['newvalue']
+                    break
+                # cc and attachments.isobsolete sometime appear
+                # in different activities with same time
+                if ((field_name == "cc" or field_name == "attachments.isobsolete") \
+                    and oldChange['time'] == ticketChange['time']):
+                    oldChange['newvalue'] += ", " + ticketChange['newvalue']
+                    break
             else:
                 ticketChanges.append (ticketChange)
 
@@ -811,26 +880,35 @@ def convert(_db, _host, _user, _password, _env, _force):
         ticket['keywords'] = string.join(keywords)
         ticketid = trac.addTicket(**ticket)
 
-        if BZ_VERSION >= 2180:
-            mysql_cur.execute("SELECT attachments.*, attach_data.thedata FROM attachments, attach_data
-                    WHERE attachments.bug_id = %s AND attachments.attach_id = attach_data.id" % bugid)
+        if BZ_VERSION >= 2210:
+            mysql_cur.execute("SELECT attachments.*, attach_data.thedata "
+                              "FROM attachments, attach_data "
+                              "WHERE attachments.bug_id = %s AND "
+                              "attachments.attach_id = attach_data.id" % bugid)
         else:
-            mysql_cur.execute("SELECT * FROM attachments WHERE bug_id = %s" % bugid)
+            mysql_cur.execute("SELECT * FROM attachments WHERE bug_id = %s" %
+                              bugid)
         attachments = mysql_cur.fetchall()
         for a in attachments:
             author = trac.getLoginName(mysql_cur, a['submitter_id'])
             trac.addAttachment(author, a)
 
     print "\n8. Importing users and passwords..."
-    if BZ_VERSION >= 2180:
-        mysql_cur.execute("SELECT login_name, cryptpassword FROM profiles")
+    if BZ_VERSION >= 2164:
+        selectlogins = "SELECT login_name, cryptpassword FROM profiles";
+        if IGNORE_DISABLED_USERS:
+            selectlogins = selectlogins + " WHERE disabledtext=''"
+        mysql_cur.execute(selectlogins)
         users = mysql_cur.fetchall()
+    else:
+        users = ()
     htpasswd = file("htpasswd", 'w')
     for user in users:
-        if LOGIN_MAP.has_key(user['login_name']):
+        if user['login_name'] in LOGIN_MAP:
             login = LOGIN_MAP[user['login_name']]
         else:
             login = user['login_name']
+
         htpasswd.write(login + ":" + user['cryptpassword'] + "\n")
 
     htpasswd.close()
@@ -841,6 +919,10 @@ def convert(_db, _host, _user, _password, _env, _force):
 def log(msg):
     print "DEBUG: %s" % (msg)
 
+def datetime2epoch(dt) :
+    import time
+    return time.mktime(dt.timetuple()) * 1000000
+
 def usage():
     print """bugzilla2trac - Imports a bug database from Bugzilla into Trac.
 
@@ -854,6 +936,8 @@ Available Options:
   -p | --passwd <MySQL password>   - Bugzilla's user password
   -c | --clean                     - Remove current Trac tickets before
                                      importing
+  -n | --noseverities              - import Bugzilla severities as Trac
+                                     priorities and forget Bugzilla priorities
   --help | help                    - This help info
 
 Additional configuration options can be defined directly in the script.
@@ -862,32 +946,38 @@ Additional configuration options can be defined directly in the script.
 
 def main():
     global BZ_DB, BZ_HOST, BZ_USER, BZ_PASSWORD, TRAC_ENV, TRAC_CLEAN
+    global SEVERITIES, PRIORITIES, PRIORITIES_MAP
     if len (sys.argv) > 1:
-    	if sys.argv[1] in ['--help','help'] or len(sys.argv) < 4:
-    	    usage()
-    	iter = 1
-    	while iter < len(sys.argv):
-    	    if sys.argv[iter] in ['--db'] and iter+1 < len(sys.argv):
-    	        BZ_DB = sys.argv[iter+1]
-    	        iter = iter + 1
-    	    elif sys.argv[iter] in ['-h', '--host'] and iter+1 < len(sys.argv):
-    	        BZ_HOST = sys.argv[iter+1]
-    	        iter = iter + 1
-    	    elif sys.argv[iter] in ['-u', '--user'] and iter+1 < len(sys.argv):
-    	        BZ_USER = sys.argv[iter+1]
-    	        iter = iter + 1
-    	    elif sys.argv[iter] in ['-p', '--passwd'] and iter+1 < len(sys.argv):
-    	        BZ_PASSWORD = sys.argv[iter+1]
-    	        iter = iter + 1
-    	    elif sys.argv[iter] in ['--tracenv'] and iter+1 < len(sys.argv):
-    	        TRAC_ENV = sys.argv[iter+1]
-    	        iter = iter + 1
-    	    elif sys.argv[iter] in ['-c', '--clean']:
-    	        TRAC_CLEAN = 1
-    	    else:
-    	        print "Error: unknown parameter: " + sys.argv[iter]
-    	        sys.exit(0)
-    	    iter = iter + 1
+        if sys.argv[1] in ['--help','help'] or len(sys.argv) < 4:
+            usage()
+        iter = 1
+        while iter < len(sys.argv):
+            if sys.argv[iter] in ['--db'] and iter+1 < len(sys.argv):
+                BZ_DB = sys.argv[iter+1]
+                iter = iter + 1
+            elif sys.argv[iter] in ['-h', '--host'] and iter+1 < len(sys.argv):
+                BZ_HOST = sys.argv[iter+1]
+                iter = iter + 1
+            elif sys.argv[iter] in ['-u', '--user'] and iter+1 < len(sys.argv):
+                BZ_USER = sys.argv[iter+1]
+                iter = iter + 1
+            elif sys.argv[iter] in ['-p', '--passwd'] and iter+1 < len(sys.argv):
+                BZ_PASSWORD = sys.argv[iter+1]
+                iter = iter + 1
+            elif sys.argv[iter] in ['--tracenv'] and iter+1 < len(sys.argv):
+                TRAC_ENV = sys.argv[iter+1]
+                iter = iter + 1
+            elif sys.argv[iter] in ['-c', '--clean']:
+                TRAC_CLEAN = 1
+            elif sys.argv[iter] in ['-n', '--noseverities']:
+                # treat Bugzilla severites as Trac priorities
+                PRIORITIES = SEVERITIES
+                SEVERITIES = []
+                PRIORITIES_MAP = {}
+            else:
+                print "Error: unknown parameter: " + sys.argv[iter]
+                sys.exit(0)
+            iter = iter + 1
 
     convert(BZ_DB, BZ_HOST, BZ_USER, BZ_PASSWORD, TRAC_ENV, TRAC_CLEAN)
 
