@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2012 Edgewall Software
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -11,4 +11,9 @@
 # individuals. For the exact contribution history, see the revision
 # history and logs, available at http://trac.edgewall.org/log/.
 
-__version__ = __import__('pkg_resources').get_distribution('Trac').version
+from pkg_resources import DistributionNotFound, get_distribution
+
+try:
+    __version__ = get_distribution('Trac').version
+except DistributionNotFound:
+    __version__ = '1.0.2'
