@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # All rights reserved.
 #
@@ -14,20 +14,19 @@
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
 
+import copy
 import re
-from datetime import datetime
 
 from genshi.builder import tag
 
+from trac.cache import cached
 from trac.config import *
 from trac.core import *
-from trac.perm import IPermissionRequestor, PermissionSystem, PermissionError
+from trac.perm import IPermissionRequestor, PermissionCache, PermissionSystem
 from trac.resource import IResourceManager
-from trac.util import Ranges
-from trac.util.compat import set, sorted
-from trac.util.datefmt import utc
-from trac.util.text import shorten_line, obfuscate_email_address
-from trac.util.translation import _
+from trac.util import Ranges, as_int
+from trac.util.text import shorten_line
+from trac.util.translation import _, N_, gettext
 from trac.wiki import IWikiSyntaxProvider, WikiParser
 
 
@@ -48,7 +47,7 @@ class ITicketActionController(Interface):
         `action` is a key used to identify that particular action.
         (note that 'history' and 'diff' are reserved and should not be used
         by plugins)
-        
+
         The actions will be presented on the page in descending order of the
         integer weight. The first action in the list is used as the default
         action.
@@ -70,7 +69,7 @@ class ITicketActionController(Interface):
         `label` is a short text that will be used when listing the action,
         `control` is the markup for the action control and `hint` should
         explain what will happen if this action is taken.
-        
+
         This method will only be called if the controller claimed to handle
         the given `action` in the call to `get_ticket_actions`.
 
@@ -115,7 +114,7 @@ class ITicketChangeListener(Interface):
 
     def ticket_changed(ticket, comment, author, old_values):
         """Called when a ticket is modified.
-        
+
         `old_values` is a dictionary containing the previous values of the
         fields that have changed.
         """
@@ -123,6 +122,15 @@ class ITicketChangeListener(Interface):
     def ticket_deleted(ticket):
         """Called when a ticket is deleted."""
 
+    def ticket_comment_modified(ticket, cdate, author, comment, old_comment):
+        """Called when a ticket comment is modified."""
+
+    def ticket_change_deleted(ticket, cdate, changes):
+        """Called when a ticket change is deleted.
+
+        `changes` is a dictionary of tuple `(oldvalue, newvalue)`
+        containing the ticket change of the fields that have changed."""
+
 
 class ITicketManipulator(Interface):
     """Miscellaneous manipulation of ticket workflow features."""
@@ -133,16 +141,41 @@ class ITicketManipulator(Interface):
 
     def validate_ticket(req, ticket):
         """Validate a ticket after it's been populated from user input.
-        
+
         Must return a list of `(field, message)` tuples, one for each problem
         detected. `field` can be `None` to indicate an overall problem with the
         ticket. Therefore, a return value of `[]` means everything is OK."""
 
 
+class IMilestoneChangeListener(Interface):
+    """Extension point interface for components that require notification
+    when milestones are created, modified, or deleted."""
+
+    def milestone_created(milestone):
+        """Called when a milestone is created."""
+
+    def milestone_changed(milestone, old_values):
+        """Called when a milestone is modified.
+
+        `old_values` is a dictionary containing the previous values of the
+        milestone properties that changed. Currently those properties can be
+        'name', 'due', 'completed', or 'description'.
+        """
+
+    def milestone_deleted(milestone):
+        """Called when a milestone is deleted."""
+
+
 class TicketSystem(Component):
     implements(IPermissionRequestor, IWikiSyntaxProvider, IResourceManager)
 
     change_listeners = ExtensionPoint(ITicketChangeListener)
+    milestone_change_listeners = ExtensionPoint(IMilestoneChangeListener)
+
+    ticket_custom_section = ConfigSection('ticket-custom',
+        """In this section, you can define additional fields for tickets. See
+        TracTicketsCustomFields for more details.""")
+
     action_controllers = OrderedExtensionsOption('ticket', 'workflow',
         ITicketActionController, default='ConfigurableTicketWorkflow',
         include_missing=False,
@@ -150,12 +183,55 @@ class TicketSystem(Component):
             (''since 0.11'').""")
 
     restrict_owner = BoolOption('ticket', 'restrict_owner', 'false',
-        """Make the owner field of tickets use a drop-down menu. See
-        [TracTickets#Assign-toasDrop-DownList Assign-to as Drop-Down List]
-        (''since 0.9'').""")
+        """Make the owner field of tickets use a drop-down menu.
+        Be sure to understand the performance implications before activating
+        this option. See
+        [TracTickets#Assign-toasDrop-DownList Assign-to as Drop-Down List].
+
+        Please note that e-mail addresses are '''not''' obfuscated in the
+        resulting drop-down menu, so this option should not be used if
+        e-mail addresses must remain protected.
+        (''since 0.9'')""")
+
+    default_version = Option('ticket', 'default_version', '',
+        """Default version for newly created tickets.""")
+
+    default_type = Option('ticket', 'default_type', 'defect',
+        """Default type for newly created tickets (''since 0.9'').""")
+
+    default_priority = Option('ticket', 'default_priority', 'major',
+        """Default priority for newly created tickets.""")
+
+    default_milestone = Option('ticket', 'default_milestone', '',
+        """Default milestone for newly created tickets.""")
+
+    default_component = Option('ticket', 'default_component', '',
+        """Default component for newly created tickets.""")
+
+    default_severity = Option('ticket', 'default_severity', '',
+        """Default severity for newly created tickets.""")
+
+    default_summary = Option('ticket', 'default_summary', '',
+        """Default summary (title) for newly created tickets.""")
+
+    default_description = Option('ticket', 'default_description', '',
+        """Default description for newly created tickets.""")
+
+    default_keywords = Option('ticket', 'default_keywords', '',
+        """Default keywords for newly created tickets.""")
+
+    default_owner = Option('ticket', 'default_owner', '< default >',
+        """Default owner for newly created tickets.""")
+
+    default_cc = Option('ticket', 'default_cc', '',
+        """Default cc: list for newly created tickets.""")
+
+    default_resolution = Option('ticket', 'default_resolution', 'fixed',
+        """Default resolution for resolving (closing) tickets
+        (''since 0.11'').""")
 
     def __init__(self):
-        self.log.debug('action controllers for ticket workflow: %r' % 
+        self.log.debug('action controllers for ticket workflow: %r' %
                 [c.__class__.__name__ for c in self.action_controllers])
 
     # Public API
@@ -165,7 +241,7 @@ class TicketSystem(Component):
         # The list should not have duplicates.
         actions = {}
         for controller in self.action_controllers:
-            weighted_actions = controller.get_ticket_actions(req, ticket)
+            weighted_actions = controller.get_ticket_actions(req, ticket) or []
             for weight, action in weighted_actions:
                 if action in actions:
                     actions[action] = max(actions[action], weight)
@@ -180,54 +256,74 @@ class TicketSystem(Component):
         controllers know about."""
         valid_states = set()
         for controller in self.action_controllers:
-            valid_states.update(controller.get_all_status())
+            valid_states.update(controller.get_all_status() or [])
         return sorted(valid_states)
 
+    def get_ticket_field_labels(self):
+        """Produce a (name,label) mapping from `get_ticket_fields`."""
+        labels = dict((f['name'], f['label'])
+                      for f in self.get_ticket_fields())
+        labels['attachment'] = _("Attachment")
+        return labels
+
     def get_ticket_fields(self):
-        """Returns the list of fields available for tickets."""
+        """Returns list of fields available for tickets.
+
+        Each field is a dict with at least the 'name', 'label' (localized)
+        and 'type' keys.
+        It may in addition contain the 'custom' key, the 'optional' and the
+        'options' keys. When present 'custom' and 'optional' are always `True`.
+        """
+        fields = copy.deepcopy(self.fields)
+        label = 'label' # workaround gettext extraction bug
+        for f in fields:
+            f[label] = gettext(f[label])
+        return fields
+
+    def reset_ticket_fields(self):
+        """Invalidate ticket field cache."""
+        del self.fields
+
+    @cached
+    def fields(self, db):
+        """Return the list of fields available for tickets."""
         from trac.ticket import model
 
-        db = self.env.get_db_cnx()
         fields = []
 
         # Basic text fields
-        for name in ('summary', 'reporter'):
-            field = {'name': name, 'type': 'text', 'label': name.title()}
-            fields.append(field)
+        fields.append({'name': 'summary', 'type': 'text',
+                       'label': N_('Summary')})
+        fields.append({'name': 'reporter', 'type': 'text',
+                       'label': N_('Reporter')})
 
-        # Owner field, can be text or drop-down depending on configuration
-        field = {'name': 'owner', 'label': 'Owner'}
-        if self.restrict_owner:
-            field['type'] = 'select'
-            perm = PermissionSystem(self.env)
-            field['options'] = perm.get_users_with_permission('TICKET_MODIFY')
-            field['options'].sort()
-            field['optional'] = True
-        else:
-            field['type'] = 'text'
+        # Owner field, by default text but can be changed dynamically
+        # into a drop-down depending on configuration (restrict_owner=true)
+        field = {'name': 'owner', 'label': N_('Owner')}
+        field['type'] = 'text'
         fields.append(field)
 
         # Description
         fields.append({'name': 'description', 'type': 'textarea',
-                       'label': 'Description'})
+                       'label': N_('Description')})
 
         # Default select and radio fields
-        selects = [('type', model.Type),
-                   ('status', model.Status),
-                   ('priority', model.Priority),
-                   ('milestone', model.Milestone),
-                   ('component', model.Component),
-                   ('version', model.Version),
-                   ('severity', model.Severity),
-                   ('resolution', model.Resolution)]
-        for name, cls in selects:
+        selects = [('type', N_('Type'), model.Type),
+                   ('status', N_('Status'), model.Status),
+                   ('priority', N_('Priority'), model.Priority),
+                   ('milestone', N_('Milestone'), model.Milestone),
+                   ('component', N_('Component'), model.Component),
+                   ('version', N_('Version'), model.Version),
+                   ('severity', N_('Severity'), model.Severity),
+                   ('resolution', N_('Resolution'), model.Resolution)]
+        for name, label, cls in selects:
             options = [val.name for val in cls.select(self.env, db=db)]
             if not options:
                 # Fields without possible values are treated as if they didn't
                 # exist
                 continue
-            field = {'name': name, 'type': 'select', 'label': name.title(),
-                     'value': self.config.get('ticket', 'default_' + name),
+            field = {'name': name, 'type': 'select', 'label': label,
+                     'value': getattr(self, 'default_' + name, ''),
                      'options': options}
             if name in ('status', 'resolution'):
                 field['type'] = 'radio'
@@ -237,15 +333,26 @@ class TicketSystem(Component):
             fields.append(field)
 
         # Advanced text fields
-        for name in ('keywords', 'cc', ):
-            field = {'name': name, 'type': 'text', 'label': name.title()}
-            fields.append(field)
+        fields.append({'name': 'keywords', 'type': 'text', 'format': 'list',
+                       'label': N_('Keywords')})
+        fields.append({'name': 'cc', 'type': 'text',  'format': 'list',
+                       'label': N_('Cc')})
+
+        # Date/time fields
+        fields.append({'name': 'time', 'type': 'time',
+                       'label': N_('Created')})
+        fields.append({'name': 'changetime', 'type': 'time',
+                       'label': N_('Modified')})
 
         for field in self.get_custom_fields():
             if field['name'] in [f['name'] for f in fields]:
                 self.log.warning('Duplicate field name "%s" (ignoring)',
                                  field['name'])
                 continue
+            if field['name'] in self.reserved_field_names:
+                self.log.warning('Field name "%s" is a reserved name '
+                                 '(ignoring)', field['name'])
+                continue
             if not re.match('^[a-zA-Z][a-zA-Z0-9_]+$', field['name']):
                 self.log.warning('Invalid name for custom field: "%s" '
                                  '(ignoring)', field['name'])
@@ -255,9 +362,18 @@ class TicketSystem(Component):
 
         return fields
 
+    reserved_field_names = ['report', 'order', 'desc', 'group', 'groupdesc',
+                            'col', 'row', 'format', 'max', 'page', 'verbose',
+                            'comment', 'or']
+
     def get_custom_fields(self):
+        return copy.deepcopy(self.custom_fields)
+
+    @cached
+    def custom_fields(self, db):
+        """Return the list of custom ticket fields available for tickets."""
         fields = []
-        config = self.config['ticket-custom']
+        config = self.ticket_custom_section
         for name in [option for option, value in config.options()
                      if '.' not in option]:
             field = {
@@ -272,23 +388,54 @@ class TicketSystem(Component):
                 if '' in field['options']:
                     field['optional'] = True
                     field['options'].remove('')
+            elif field['type'] == 'text':
+                field['format'] = config.get(name + '.format', 'plain')
             elif field['type'] == 'textarea':
+                field['format'] = config.get(name + '.format', 'plain')
                 field['width'] = config.getint(name + '.cols')
                 field['height'] = config.getint(name + '.rows')
             fields.append(field)
 
-        fields.sort(lambda x, y: cmp(x['order'], y['order']))
+        fields.sort(lambda x, y: cmp((x['order'], x['name']),
+                                     (y['order'], y['name'])))
         return fields
 
+    def get_field_synonyms(self):
+        """Return a mapping from field name synonyms to field names.
+        The synonyms are supposed to be more intuitive for custom queries."""
+        # i18n TODO - translated keys
+        return {'created': 'time', 'modified': 'changetime'}
+
+    def eventually_restrict_owner(self, field, ticket=None):
+        """Restrict given owner field to be a list of users having
+        the TICKET_MODIFY permission (for the given ticket)
+        """
+        if self.restrict_owner:
+            field['type'] = 'select'
+            possible_owners = []
+            for user in PermissionSystem(self.env) \
+                    .get_users_with_permission('TICKET_MODIFY'):
+                if not ticket or \
+                        'TICKET_MODIFY' in PermissionCache(self.env, user,
+                                                           ticket.resource):
+                    possible_owners.append(user)
+            possible_owners.sort()
+            possible_owners.insert(0, '< default >')
+            field['options'] = possible_owners
+            field['optional'] = True
+
     # IPermissionRequestor methods
 
     def get_permission_actions(self):
         return ['TICKET_APPEND', 'TICKET_CREATE', 'TICKET_CHGPROP',
                 'TICKET_VIEW', 'TICKET_EDIT_CC', 'TICKET_EDIT_DESCRIPTION',
+                'TICKET_EDIT_COMMENT', 'TICKET_BATCH_MODIFY',
                 ('TICKET_MODIFY', ['TICKET_APPEND', 'TICKET_CHGPROP']),
                 ('TICKET_ADMIN', ['TICKET_CREATE', 'TICKET_MODIFY',
                                   'TICKET_VIEW', 'TICKET_EDIT_CC',
-                                  'TICKET_EDIT_DESCRIPTION'])]
+                                  'TICKET_EDIT_DESCRIPTION',
+                                  'TICKET_EDIT_COMMENT',
+                                  'TICKET_BATCH_MODIFY'])]
 
     # IWikiSyntaxProvider methods
 
@@ -318,27 +465,28 @@ class TicketSystem(Component):
                 num = r.a
                 ticket = formatter.resource('ticket', num)
                 from trac.ticket.model import Ticket
-                if Ticket.id_is_valid(num):
-                    # TODO: watch #6436 and when done, attempt to retrieve 
-                    #       ticket directly (try: Ticket(self.env, num) ...)
-                    cursor = formatter.db.cursor() 
-                    cursor.execute("SELECT type,summary,status,resolution "
-                                   "FROM ticket WHERE id=%s", (str(num),)) 
-                    for type, summary, status, resolution in cursor:
+                if Ticket.id_is_valid(num) and \
+                        'TICKET_VIEW' in formatter.perm(ticket):
+                    # TODO: attempt to retrieve ticket view directly,
+                    #       something like: t = Ticket.view(num)
+                    for type, summary, status, resolution in \
+                            self.env.db_query("""
+                            SELECT type, summary, status, resolution
+                            FROM ticket WHERE id=%s
+                            """, (str(num),)):
                         title = self.format_summary(summary, status,
                                                     resolution, type)
                         href = formatter.href.ticket(num) + params + fragment
-                        return tag.a(label, class_='%s ticket' % status, 
-                                     title=title, href=href)
-                    else:
-                        href = formatter.href.ticket(num)
-                        return tag.a(label, class_='missing ticket', href=href,
-                                     rel="nofollow")
+                        return tag.a(label, title=title, href=href,
+                                     class_='%s ticket' % status)
             else:
                 ranges = str(r)
                 if params:
                     params = '&' + params[1:]
-                return tag.a(label, title='Tickets '+ranges,
+                label_wrap = label.replace(',', u',\u200b')
+                ranges_wrap = ranges.replace(',', u', ')
+                return tag.a(label_wrap,
+                             title=_("Tickets %(ranges)s", ranges=ranges_wrap),
                              href=formatter.href.query(id=ranges) + params)
         except ValueError:
             pass
@@ -352,19 +500,46 @@ class TicketSystem(Component):
                 cnum, realm, id = elts
                 if cnum != 'description' and cnum and not cnum[0].isdigit():
                     realm, id, cnum = elts # support old comment: style
+                id = as_int(id, None)
                 resource = formatter.resource(realm, id)
         else:
             resource = formatter.resource
             cnum = target
 
-        if resource:
-            href = "%s#comment:%s" % (formatter.href.ticket(resource.id), cnum)
-            title = _("Comment %(cnum)s for Ticket #%(id)s", cnum=cnum,
-                      id=resource.id)
-            return tag.a(label, href=href, title=title)
-        else:
-            return label
- 
+        if resource and resource.id and resource.realm == 'ticket' and \
+                cnum and (all(c.isdigit() for c in cnum) or cnum == 'description'):
+            href = title = class_ = None
+            if self.resource_exists(resource):
+                from trac.ticket.model import Ticket
+                ticket = Ticket(self.env, resource.id)
+                if cnum != 'description' and not ticket.get_change(cnum):
+                    title = _("ticket comment does not exist")
+                    class_ = 'missing ticket'
+                elif 'TICKET_VIEW' in formatter.perm(resource):
+                    href = formatter.href.ticket(resource.id) + \
+                           "#comment:%s" % cnum
+                    if resource.id != formatter.resource.id:
+                        if cnum == 'description':
+                            title = _("Description for Ticket #%(id)s",
+                                      id=resource.id)
+                        else:
+                            title = _("Comment %(cnum)s for Ticket #%(id)s",
+                                      cnum=cnum, id=resource.id)
+                        class_ = ticket['status'] + ' ticket'
+                    else:
+                        title = _("Description") if cnum == 'description' \
+                                                 else _("Comment %(cnum)s",
+                                                        cnum=cnum)
+                        class_ = 'ticket'
+                else:
+                    title = _("no permission to view ticket")
+                    class_ = 'forbidden ticket'
+            else:
+                title = _("ticket does not exist")
+                class_ = 'missing ticket'
+            return tag.a(label, class_=class_, href=href, title=title)
+        return label
+
     # IResourceManager methods
 
     def get_resource_realms(self):
@@ -392,3 +567,30 @@ class TicketSystem(Component):
             return "%s (%s)" % (summary, status)
         else:
             return summary
+
+    def resource_exists(self, resource):
+        """
+        >>> from trac.test import EnvironmentStub
+        >>> from trac.resource import Resource, resource_exists
+        >>> env = EnvironmentStub()
+
+        >>> resource_exists(env, Resource('ticket', 123456))
+        False
+
+        >>> from trac.ticket.model import Ticket
+        >>> t = Ticket(env)
+        >>> int(t.insert())
+        1
+        >>> resource_exists(env, t.resource)
+        True
+        """
+        if self.env.db_query("SELECT id FROM ticket WHERE id=%s",
+                             (resource.id,)):
+            if resource.version is None:
+                return True
+            revcount = self.env.db_query("""
+                SELECT count(DISTINCT time) FROM ticket_change WHERE ticket=%s
+                """, (resource.id,))
+            return revcount[0][0] >= resource.version
+        else:
+            return False
