@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2006-2008 Edgewall Software
+# Copyright (C) 2006-2009 Edgewall Software
 # Copyright (C) 2006 Alec Thomas
 # Copyright (C) 2007 Eli Carter
-# Copyright (C) 2007 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2007 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -18,15 +18,22 @@
 
 import pkg_resources
 
+from ConfigParser import RawConfigParser
+from StringIO import StringIO
+from functools import partial
+
 from genshi.builder import tag
 
+from trac.config import Configuration, ConfigSection
 from trac.core import *
-from trac.perm import PermissionSystem
 from trac.env import IEnvironmentSetupParticipant
-from trac.config import Configuration
+from trac.perm import PermissionSystem
 from trac.ticket.api import ITicketActionController, TicketSystem
-from trac.util.compat import set
-from trac.util.translation import _
+from trac.ticket.model import Resolution
+from trac.util.text import obfuscate_email_address
+from trac.util.translation import _, tag_, cleandoc_
+from trac.web.chrome import Chrome, add_script, add_script_data
+from trac.wiki.macros import WikiMacroBase
 
 # -- Utilities for the ConfigurableTicketWorkflow
 
@@ -37,13 +44,13 @@ def parse_workflow_config(rawactions):
         parts = option.split('.')
         action = parts[0]
         if action not in actions:
-            actions[action] = {}
+            actions[action] = {'oldstates': '', 'newstate': ''}
         if len(parts) == 1:
             # Base name, of the syntax: old,states,here -> newstate
             try:
                 oldstates, newstate = [x.strip() for x in value.split('->')]
             except ValueError:
-                raise Exception('Bad option "%s"' % (option, )) # 500, no _
+                continue # Syntax error, a warning will be logged later
             actions[action]['newstate'] = newstate
             actions[action]['oldstates'] = oldstates
         else:
@@ -51,30 +58,22 @@ def parse_workflow_config(rawactions):
             actions[action][attribute] = value
     # Fill in the defaults for every action, and normalize them to the desired
     # types
+    def as_list(key):
+        value = attributes.get(key, '')
+        return [item for item in (x.strip() for x in value.split(',')) if item]
+
     for action, attributes in actions.items():
         # Default the 'name' attribute to the name used in the ini file
         if 'name' not in attributes:
             attributes['name'] = action
         # If not specified, an action is not the default.
-        if 'default' not in attributes:
-            attributes['default'] = 0
-        else:
-            attributes['default'] = int(attributes['default'])
+        attributes['default'] = int(attributes.get('default', 0))
         # If operations are not specified, that means no operations
-        if 'operations' not in attributes:
-            attributes['operations'] = []
-        else:
-            attributes['operations'] = [a.strip() for a in
-                                        attributes['operations'].split(',')]
+        attributes['operations'] = as_list('operations')
         # If no permissions are specified, then no permissions are needed
-        if 'permissions' not in attributes:
-            attributes['permissions'] = []
-        else:
-            attributes['permissions'] = [a.strip() for a in
-                                         attributes['permissions'].split(',')]
+        attributes['permissions'] = as_list('permissions')
         # Normalize the oldstates
-        attributes['oldstates'] = [x.strip() for x in
-                                   attributes['oldstates'].split(',')]
+        attributes['oldstates'] = as_list('oldstates')
     return actions
 
 def get_workflow_config(config):
@@ -98,12 +97,23 @@ def load_workflow_config_snippet(config, filename):
 
 class ConfigurableTicketWorkflow(Component):
     """Ticket action controller which provides actions according to a
-    workflow defined in the TracIni configuration file, inside the
-    [ticket-workflow] section.
+    workflow defined in trac.ini.
+
+    The workflow is defined in the `[ticket-workflow]` section of the
+    [wiki:TracIni#ticket-workflow-section trac.ini] configuration file.
     """
-    
+
+    implements(IEnvironmentSetupParticipant, ITicketActionController)
+
+    ticket_workflow_section = ConfigSection('ticket-workflow',
+        """The workflow for tickets is controlled by plugins. By default,
+        there's only a `ConfigurableTicketWorkflow` component in charge.
+        That component allows the workflow to be configured via this section
+        in the `trac.ini` file. See TracWorkflow for more details.
+
+        (''since 0.11'')""")
+
     def __init__(self, *args, **kwargs):
-        Component.__init__(self, *args, **kwargs)
         self.actions = get_workflow_config(self.config)
         if not '_reset' in self.actions:
             # Special action that gets enabled if the current status no longer
@@ -117,8 +127,11 @@ class ConfigurableTicketWorkflow(Component):
                 'permissions': []}
         self.log.debug('Workflow actions at initialization: %s\n' %
                        str(self.actions))
+        for name, info in self.actions.iteritems():
+            if not info['newstate']:
+                self.log.warning("Ticket workflow action '%s' doesn't define "
+                                 "any transitions", name)
 
-    implements(ITicketActionController, IEnvironmentSetupParticipant)
 
     # IEnvironmentSetupParticipant methods
 
@@ -150,7 +163,7 @@ The ticket Workflow is now configurable.
 
 Your environment has been upgraded, but configured to use the original
 workflow. It is recommended that you look at changing this configuration to use
-basic-workflow. 
+basic-workflow.
 
 Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
 
@@ -169,36 +182,33 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         # the process of being modified, we need to base our information on the
         # pre-modified state so that we don't try to do two (or more!) steps at
         # once and get really confused.
-        if 'status' in ticket._old:
-            status = ticket._old['status']
-        else:
-            status = ticket['status']
-        status = status or 'new'
+        status = ticket._old.get('status', ticket['status']) or 'new'
 
+        ticket_perm = req.perm(ticket.resource)
         allowed_actions = []
         for action_name, action_info in self.actions.items():
             oldstates = action_info['oldstates']
             if oldstates == ['*'] or status in oldstates:
                 # This action is valid in this state.  Check permissions.
-                allowed = 0
                 required_perms = action_info['permissions']
-                if required_perms:
-                    for permission in required_perms:
-                        if permission in req.perm(ticket.resource):
-                            allowed = 1
-                            break
-                else:
-                    allowed = 1
-                if allowed:
+                if self._is_action_allowed(ticket_perm, required_perms):
                     allowed_actions.append((action_info['default'],
                                             action_name))
         if not (status in ['new', 'closed'] or \
                     status in TicketSystem(self.env).get_all_status()) \
-                and 'TICKET_ADMIN' in req.perm(ticket.resource):
+                and 'TICKET_ADMIN' in ticket_perm:
             # State no longer exists - add a 'reset' action if admin.
             allowed_actions.append((0, '_reset'))
         return allowed_actions
 
+    def _is_action_allowed(self, ticket_perm, required_perms):
+        if not required_perms:
+            return True
+        for permission in required_perms:
+            if permission in ticket_perm:
+                return True
+        return False
+
     def get_all_status(self):
         """Return a list of all states described by the configuration.
 
@@ -208,21 +218,24 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
             all_status.update(action_info['oldstates'])
             all_status.add(action_info['newstate'])
         all_status.discard('*')
+        all_status.discard('')
         return all_status
-        
+
     def render_ticket_action_control(self, req, ticket, action):
-        from trac.ticket import model
 
         self.log.debug('render_ticket_action_control: action "%s"' % action)
 
         this_action = self.actions[action]
-        status = this_action['newstate']        
+        status = this_action['newstate']
         operations = this_action['operations']
+        current_owner = ticket._old.get('owner', ticket['owner'])
+        format_author = partial(Chrome(self.env).format_author, req)
+        formatted_current_owner = format_author(current_owner or _("(none)"))
 
         control = [] # default to nothing
         hints = []
         if 'reset_workflow' in operations:
-            control.append(tag("from invalid state "))
+            control.append(_("from invalid state"))
             hints.append(_("Current state no longer exists"))
         if 'del_owner' in operations:
             hints.append(_("The ticket will be disowned"))
@@ -230,7 +243,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
             id = 'action_%s_reassign_owner' % action
             selected_owner = req.args.get(id, req.authname)
 
-            if this_action.has_key('set_owner'):
+            if 'set_owner' in this_action:
                 owners = [x.strip() for x in
                           this_action['set_owner'].split(',')]
             elif self.config.getbool('ticket', 'restrict_owner'):
@@ -240,54 +253,86 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
             else:
                 owners = None
 
-            if owners == None:
+            if owners is None:
                 owner = req.args.get(id, req.authname)
-                control.append(tag(['to ', tag.input(type='text', id=id,
-                                                     name=id, value=owner)]))
-                hints.append(_("The owner will change"))
+                control.append(tag_("to %(owner)s",
+                                    owner=tag.input(type='text', id=id,
+                                                    name=id, value=owner)))
+                hints.append(_("The owner will be changed from "
+                               "%(current_owner)s to the specified user",
+                               current_owner=formatted_current_owner))
             elif len(owners) == 1:
-                control.append(tag('to %s ' % owners[0]))
+                owner = tag.input(type='hidden', id=id, name=id,
+                                  value=owners[0])
+                formatted_new_owner = format_author(owners[0])
+                control.append(tag_("to %(owner)s",
+                                    owner=tag(formatted_new_owner, owner)))
                 if ticket['owner'] != owners[0]:
-                    hints.append(_("The owner will change to %s") % owners[0])
+                    hints.append(_("The owner will be changed from "
+                                   "%(current_owner)s to %(selected_owner)s",
+                                   current_owner=formatted_current_owner,
+                                   selected_owner=formatted_new_owner))
             else:
-                control.append(tag([_("to "), tag.select(
-                    [tag.option(x, selected=(x == selected_owner or None))
+                control.append(tag_("to %(owner)s", owner=tag.select(
+                    [tag.option(x, value=x,
+                                selected=(x == selected_owner or None))
                      for x in owners],
-                    id=id, name=id)]))
-                hints.append(_("The owner will change"))
-        if 'set_owner_to_self' in operations and \
-            ticket['owner'] != req.authname:
-            hints.append(_("The owner will change to %s") % req.authname)
+                    id=id, name=id)))
+                hints.append(_("The owner will be changed from "
+                               "%(current_owner)s to the selected user",
+                               current_owner=formatted_current_owner))
+        elif 'set_owner_to_self' in operations and \
+                ticket._old.get('owner', ticket['owner']) != req.authname:
+            hints.append(_("The owner will be changed from %(current_owner)s "
+                           "to %(authname)s",
+                           current_owner=formatted_current_owner,
+                           authname=format_author(req.authname)))
         if 'set_resolution' in operations:
-            if this_action.has_key('set_resolution'):
+            if 'set_resolution' in this_action:
                 resolutions = [x.strip() for x in
                                this_action['set_resolution'].split(',')]
             else:
-                resolutions = [val.name for val in
-                               model.Resolution.select(self.env)]
+                resolutions = [val.name for val in Resolution.select(self.env)]
             if not resolutions:
                 raise TracError(_("Your workflow attempts to set a resolution "
                                   "but none is defined (configuration issue, "
                                   "please contact your Trac admin)."))
+            id = 'action_%s_resolve_resolution' % action
             if len(resolutions) == 1:
-                control.append(tag('as %s' % resolutions[0]))
-                hints.append(_("The resolution will be set to %s") %
-                             resolutions[0])
+                resolution = tag.input(type='hidden', id=id, name=id,
+                                       value=resolutions[0])
+                control.append(tag_("as %(resolution)s",
+                                    resolution=tag(resolutions[0],
+                                                   resolution)))
+                hints.append(_("The resolution will be set to %(name)s",
+                               name=resolutions[0]))
             else:
-                id = 'action_%s_resolve_resolution' % action
                 selected_option = req.args.get(id,
-                        self.config.get('ticket', 'default_resolution'))
-                control.append(tag(['as ', tag.select(
-                    [tag.option(x, selected=(x == selected_option or None))
+                        TicketSystem(self.env).default_resolution)
+                control.append(tag_("as %(resolution)s",
+                                    resolution=tag.select(
+                    [tag.option(x, value=x,
+                                selected=(x == selected_option or None))
                      for x in resolutions],
-                    id=id, name=id)]))
+                    id=id, name=id)))
                 hints.append(_("The resolution will be set"))
+        if 'del_resolution' in operations:
+            hints.append(_("The resolution will be deleted"))
         if 'leave_status' in operations:
-            control.append('as %s ' % ticket['status'])
+            control.append(_("as %(status)s",
+                             status= ticket._old.get('status',
+                                                     ticket['status'])))
+            if len(operations) == 1:
+                hints.append(_("The owner will remain %(current_owner)s",
+                               current_owner=formatted_current_owner)
+                             if current_owner else
+                             _("The ticket will remain with no owner"))
         else:
             if status != '*':
-                hints.append(_("Next status will be '%s'") % status)
-        return (this_action['name'], tag(*control), '. '.join(hints))
+                hints.append(_("Next status will be '%(name)s'", name=status))
+        return (this_action['name'],
+                tag((' ' if i else None, c) for i, c in enumerate(control)),
+                '. '.join(hints) + '.' if hints else '')
 
     def get_ticket_changes(self, req, ticket, action):
         this_action = self.actions[action]
@@ -307,7 +352,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         for operation in this_action['operations']:
             if operation == 'reset_workflow':
                 updated['status'] = 'new'
-            if operation == 'del_owner':
+            elif operation == 'del_owner':
                 updated['owner'] = ''
             elif operation == 'set_owner':
                 newowner = req.args.get('action_%s_reassign_owner' % action,
@@ -319,8 +364,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
                 updated['owner'] = newowner
             elif operation == 'set_owner_to_self':
                 updated['owner'] = req.authname
-
-            if operation == 'del_resolution':
+            elif operation == 'del_resolution':
                 updated['resolution'] = ''
             elif operation == 'set_resolution':
                 newresolution = req.args.get('action_%s_resolve_resolution' % \
@@ -347,7 +391,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
 
     # Public methods (for other ITicketActionControllers that want to use
     #                 our config file and provide an operation for an action)
-    
+
     def get_actions_by_operation(self, operation):
         """Return a list of all actions with a given operation
         (for use in the controller's get_all_status())
@@ -374,3 +418,84 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
                       self._has_perms_for_action(req, info, ticket.resource)]
         return actions
 
+
+class WorkflowMacro(WikiMacroBase):
+    _domain = 'messages'
+    _description = cleandoc_(
+    """Render a workflow graph.
+
+    This macro accepts a TracWorkflow configuration and renders the states
+    and transitions as a directed graph. If no parameters are given, the
+    current ticket workflow is rendered. In WikiProcessors mode the `width`
+    and `height` arguments can be specified.
+
+    (Defaults: `width = 800` and `heigth = 600`)
+
+    Examples:
+    {{{
+        [[Workflow()]]
+
+        [[Workflow(go = here -> there; return = there -> here)]]
+
+        {{{
+        #!Workflow width=700 height=700
+        leave = * -> *
+        leave.operations = leave_status
+        leave.default = 1
+
+        accept = new,assigned,accepted,reopened -> accepted
+        accept.permissions = TICKET_MODIFY
+        accept.operations = set_owner_to_self
+
+        resolve = new,assigned,accepted,reopened -> closed
+        resolve.permissions = TICKET_MODIFY
+        resolve.operations = set_resolution
+
+        reassign = new,assigned,accepted,reopened -> assigned
+        reassign.permissions = TICKET_MODIFY
+        reassign.operations = set_owner
+
+        reopen = closed -> reopened
+        reopen.permissions = TICKET_CREATE
+        reopen.operations = del_resolution
+        }}}
+    }}}
+    """)
+
+    def expand_macro(self, formatter, name, text, args):
+        if not text:
+            raw_actions = self.config.options('ticket-workflow')
+        else:
+            if args is None:
+                text = '\n'.join([line.lstrip() for line in text.split(';')])
+            if not '[ticket-workflow]' in text:
+                text = '[ticket-workflow]\n' + text
+            parser = RawConfigParser()
+            parser.readfp(StringIO(text))
+            raw_actions = list(parser.items('ticket-workflow'))
+        actions = parse_workflow_config(raw_actions)
+        states = list(set(
+            [state for action in actions.itervalues()
+                   for state in action['oldstates']] +
+            [action['newstate'] for action in actions.itervalues()]))
+        action_names = actions.keys()
+        edges = []
+        for name, action in actions.items():
+            new_index = states.index(action['newstate'])
+            name_index = action_names.index(name)
+            for old_state in action['oldstates']:
+                old_index = states.index(old_state)
+                edges.append((old_index, new_index, name_index))
+
+        args = args or {}
+        graph = {'nodes': states, 'actions': action_names, 'edges': edges,
+                 'width': args.get('width', 800),
+                 'height': args.get('height', 600)}
+        graph_id = '%012x' % id(graph)
+        req = formatter.req
+        add_script(req, 'common/js/excanvas.js', ie_if='IE')
+        add_script(req, 'common/js/workflow_graph.js')
+        add_script_data(req, {'graph_%s' % graph_id: graph})
+        return tag.div(_("Enable JavaScript to display the workflow graph."),
+                       class_='trac-workflow-graph system-message',
+                       id='trac-workflow-graph-%s' % graph_id)
