@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
 #
@@ -18,19 +18,114 @@ from BaseHTTPServer import BaseHTTPRequestHandler
 from Cookie import CookieError, BaseCookie, SimpleCookie
 import cgi
 from datetime import datetime
+import errno
+from hashlib import md5
 import new
 import mimetypes
 import os
+import re
+import socket
 from StringIO import StringIO
 import sys
 import urlparse
 
+from genshi.builder import Fragment
 from trac.core import Interface, TracError
-from trac.util import get_last_traceback, md5
+from trac.perm import PermissionError
+from trac.util import get_last_traceback, unquote
 from trac.util.datefmt import http_date, localtz
+from trac.util.text import empty, exception_to_unicode, to_unicode
+from trac.util.translation import _
 from trac.web.href import Href
 from trac.web.wsgi import _FileWrapper
 
+
+class IAuthenticator(Interface):
+    """Extension point interface for components that can provide the name
+    of the remote user."""
+
+    def authenticate(req):
+        """Return the name of the remote user, or `None` if the identity of the
+        user is unknown."""
+
+
+class IRequestHandler(Interface):
+    """Decide which `trac.core.Component` handles which `Request`, and how."""
+
+    def match_request(req):
+        """Return whether the handler wants to process the given request."""
+
+    def process_request(req):
+        """Process the request.
+
+        Return a `(template_name, data, content_type)` tuple,
+        where `data` is a dictionary of substitutions for the Genshi template.
+
+        "text/html" is assumed if `content_type` is `None`.
+
+        Note that if template processing should not occur, this method can
+        simply send the response itself and not return anything.
+
+        :Since 1.0: Clearsilver templates are no longer supported.
+        """
+
+
+class IRequestFilter(Interface):
+    """Enable components to interfere with the processing done by the
+    main handler, either before and/or after it enters in action.
+    """
+
+    def pre_process_request(req, handler):
+        """Called after initial handler selection, and can be used to change
+        the selected handler or redirect request.
+
+        Always returns the request handler, even if unchanged.
+        """
+
+    def post_process_request(req, template, data, content_type):
+        """Do any post-processing the request might need; typically adding
+        values to the template `data` dictionary, or changing the Genshi
+        template or mime type.
+
+        `data` may be updated in place.
+
+        Always returns a tuple of (template, data, content_type), even if
+        unchanged.
+
+        Note that `template`, `data`, `content_type` will be `None` if:
+         - called when processing an error page
+         - the default request handler did not return any result
+
+        :Since 0.11: there's a `data` argument for supporting Genshi templates;
+           this introduced a difference in arity which made it possible to
+           distinguish between the IRequestFilter components still targeted
+           at ClearSilver templates and the newer ones targeted at Genshi
+           templates.
+
+        :Since 1.0: Clearsilver templates are no longer supported.
+        """
+
+
+class ITemplateStreamFilter(Interface):
+    """Transform the generated content by filtering the Genshi event stream
+    generated by the template, prior to its serialization.
+    """
+
+    def filter_stream(req, method, filename, stream, data):
+        """Return a filtered Genshi event stream, or the original unfiltered
+        stream if no match.
+
+        `req` is the current request object, `method` is the Genshi render
+        method (xml, xhtml or text), `filename` is the filename of the template
+        to be rendered, `stream` is the event stream and `data` is the data for
+        the current template.
+
+        See the Genshi_ documentation for more information.
+
+        .. _Genshi: http://genshi.edgewall.org/wiki/Documentation/filters.html
+        """
+
+
 HTTP_STATUS = dict([(code, reason.title()) for code, (reason, description)
                     in BaseHTTPRequestHandler.responses.items()])
 
@@ -38,7 +133,7 @@ HTTP_STATUS = dict([(code, reason.title()) for code, (reason, description)
 class HTTPException(Exception):
 
     def __init__(self, detail, *args):
-        if isinstance(detail, TracError):
+        if isinstance(detail, (TracError, PermissionError)):
             self.detail = detail.message
             self.reason = detail.title
         else:
@@ -47,6 +142,37 @@ class HTTPException(Exception):
             self.detail = self.detail % args
         Exception.__init__(self, '%s %s (%s)' % (self.code, self.reason,
                                                  self.detail))
+
+    @property
+    def message(self):
+        # The message is based on the e.detail, which can be an Exception
+        # object, but not a TracError one: when creating HTTPException,
+        # a TracError.message is directly assigned to e.detail
+        if isinstance(self.detail, Exception): # not a TracError or PermissionError
+            message = exception_to_unicode(self.detail)
+        elif isinstance(self.detail, Fragment): # TracError or PermissionError markup
+            message = self.detail
+        else:
+            message = to_unicode(self.detail)
+        return message
+
+    @property
+    def title(self):
+        try:
+            # We first try to get localized error messages here, but we
+            # should ignore secondary errors if the main error was also
+            # due to i18n issues
+            title = _("Error")
+            if self.reason:
+                if title.lower() in self.reason.lower():
+                    title = self.reason
+                else:
+                    title = _("Error: %(message)s", message=self.reason)
+        except Exception:
+            title = "Error"
+        return title
+
+    @classmethod
     def subclass(cls, name, code):
         """Create a new Exception class representing a HTTP status code."""
         reason = HTTP_STATUS.get(code, 'Unknown')
@@ -56,9 +182,8 @@ class HTTPException(Exception):
         new_class.code = code
         new_class.reason = reason
         return new_class
-    subclass = classmethod(subclass)
-
 
+_HTTPException_subclass_names = []
 for code in [code for code in HTTP_STATUS if code >= 400]:
     exc_name = HTTP_STATUS[code].replace(' ', '').replace('-', '')
     # 2.5 compatibility hack:
@@ -66,12 +191,25 @@ for code in [code for code in HTTP_STATUS if code >= 400]:
         exc_name = 'InternalError'
     if exc_name.lower().startswith('http'):
         exc_name = exc_name[4:]
-    exc_name = 'HTTP' + exc_name        
+    exc_name = 'HTTP' + exc_name
     setattr(sys.modules[__name__], exc_name,
             HTTPException.subclass(exc_name, code))
+    _HTTPException_subclass_names.append(exc_name)
 del code, exc_name
 
 
+class _FieldStorage(cgi.FieldStorage):
+    """Our own version of cgi.FieldStorage, with tweaks."""
+
+    def read_multi(self, *args, **kwargs):
+        try:
+            cgi.FieldStorage.read_multi(self, *args, **kwargs)
+        except ValueError:
+            # Most likely "Invalid boundary in multipart form",
+            # possibly an upload of a .mht file? See #9880.
+            self.read_single()
+
+
 class _RequestArgs(dict):
     """Dictionary subclass that provides convenient access to request
     parameters that may contain multiple values."""
@@ -99,10 +237,49 @@ class _RequestArgs(dict):
         return val
 
 
+def parse_arg_list(query_string):
+    """Parse a query string into a list of `(name, value)` tuples."""
+    args = []
+    if not query_string:
+        return args
+    for arg in query_string.split('&'):
+        nv = arg.split('=', 1)
+        if len(nv) == 2:
+            (name, value) = nv
+        else:
+            (name, value) = (nv[0], empty)
+        name = unquote(name.replace('+', ' '))
+        if isinstance(name, str):
+            name = unicode(name, 'utf-8')
+        value = unquote(value.replace('+', ' '))
+        if isinstance(value, str):
+            value = unicode(value, 'utf-8')
+        args.append((name, value))
+    return args
+
+
+def arg_list_to_args(arg_list):
+    """Convert a list of `(name, value)` tuples into into a `_RequestArgs`."""
+    args = _RequestArgs()
+    for name, value in arg_list:
+        if name in args:
+            if isinstance(args[name], list):
+                args[name].append(value)
+            else:
+                args[name] = [args[name], value]
+        else:
+            args[name] = value
+    return args
+
+
 class RequestDone(Exception):
     """Marker exception that indicates whether request processing has completed
     and a response was sent.
     """
+    iterable = None
+
+    def __init__(self, iterable=None):
+        self.iterable = iterable
 
 
 class Cookie(SimpleCookie):
@@ -119,6 +296,10 @@ class Cookie(SimpleCookie):
     _strict_set = BaseCookie._BaseCookie__set
 
     def _loose_set(self, key, real_value, coded_value):
+        # If a key appears multiple times, the first occurrence has the
+        # narrowest scope, keep that
+        if key in self:
+            return
         try:
             self._strict_set(key, real_value, coded_value)
         except CookieError:
@@ -128,16 +309,16 @@ class Cookie(SimpleCookie):
 
 class Request(object):
     """Represents a HTTP request/response pair.
-    
+
     This class provides a convenience API over WSGI.
     """
 
     def __init__(self, environ, start_response):
         """Create the request wrapper.
-        
-        @param environ: The WSGI environment dict
-        @param start_response: The WSGI callback for starting the response
-        @param callbacks: A dictionary of functions that are used to lazily
+
+        :param environ: The WSGI environment dict
+        :param start_response: The WSGI callback for starting the response
+        :param callbacks: A dictionary of functions that are used to lazily
             evaluate attribute lookups
         """
         self.environ = environ
@@ -151,10 +332,13 @@ class Request(object):
         self.outcookie = Cookie()
 
         self.callbacks = {
-            'args': Request._parse_args,
+            'arg_list': Request._parse_arg_list,
+            'args': lambda req: arg_list_to_args(req.arg_list),
+            'languages': Request._parse_languages,
             'incookie': Request._parse_cookies,
             '_inheaders': Request._parse_headers
         }
+        self.redirect_listeners = []
 
         self.base_url = self.environ.get('trac.base_url')
         if not self.base_url:
@@ -172,28 +356,73 @@ class Request(object):
         raise AttributeError(name)
 
     def __repr__(self):
+        path_info = self.environ.get('PATH_INFO', '')
         return '<%s "%s %r">' % (self.__class__.__name__, self.method,
-                                 self.path_info)
+                                 path_info)
 
     # Public API
 
-    method = property(fget=lambda self: self.environ['REQUEST_METHOD'],
-                      doc='The HTTP method of the request')
-    path_info = property(fget=lambda self: self.environ.get('PATH_INFO', '').decode('utf-8'),
-                         doc='Path inside the application')
-    remote_addr = property(fget=lambda self: self.environ.get('REMOTE_ADDR'),
-                           doc='IP address of the remote user')
-    remote_user = property(fget=lambda self: self.environ.get('REMOTE_USER'),
-                           doc='Name of the remote user, `None` if the user'
-                               'has not logged in using HTTP authentication')
-    scheme = property(fget=lambda self: self.environ['wsgi.url_scheme'],
-                      doc='The scheme of the request URL')
-    base_path = property(fget=lambda self: self.environ.get('SCRIPT_NAME', ''),
-                         doc='The root path of the application')
-    server_name = property(fget=lambda self: self.environ['SERVER_NAME'],
-                           doc='Name of the server')
-    server_port = property(fget=lambda self: int(self.environ['SERVER_PORT']),
-                           doc='Port number the server is bound to')
+    @property
+    def method(self):
+        """The HTTP method of the request"""
+        return self.environ['REQUEST_METHOD']
+
+    @property
+    def path_info(self):
+        """Path inside the application"""
+        path_info = self.environ.get('PATH_INFO', '')
+        try:
+            return unicode(path_info, 'utf-8')
+        except UnicodeDecodeError:
+            raise HTTPNotFound(_("Invalid URL encoding (was %(path_info)r)",
+                                 path_info=path_info))
+
+    @property
+    def query_string(self):
+        """Query part of the request"""
+        return self.environ.get('QUERY_STRING', '')
+
+    @property
+    def remote_addr(self):
+        """IP address of the remote user"""
+        return self.environ.get('REMOTE_ADDR')
+
+    @property
+    def remote_user(self):
+        """ Name of the remote user.
+
+        Will be `None` if the user has not logged in using HTTP authentication.
+        """
+        user = self.environ.get('REMOTE_USER')
+        if user is not None:
+            return to_unicode(user)
+
+    @property
+    def scheme(self):
+        """The scheme of the request URL"""
+        return self.environ['wsgi.url_scheme']
+
+    @property
+    def base_path(self):
+        """The root path of the application"""
+        return self.environ.get('SCRIPT_NAME', '')
+
+    @property
+    def server_name(self):
+        """Name of the server"""
+        return self.environ['SERVER_NAME']
+
+    @property
+    def server_port(self):
+        """Port number the server is bound to"""
+        return int(self.environ['SERVER_PORT'])
+
+    def add_redirect_listener(self, listener):
+        """Add a callable to be called prior to executing a redirect.
+
+        The callable is passed the arguments to the `redirect()` call.
+        """
+        self.redirect_listeners.append(listener)
 
     def get_header(self, name):
         """Return the value of the specified HTTP header, or `None` if there's
@@ -215,15 +444,18 @@ class Request(object):
         `value` must either be an `unicode` string or can be converted to one
         (e.g. numbers, ...)
         """
-        if name.lower() == 'content-type':
+        lower_name = name.lower()
+        if lower_name == 'content-type':
             ctpos = value.find('charset=')
             if ctpos >= 0:
                 self._outcharset = value[ctpos + 8:].strip()
+        elif lower_name == 'content-length':
+            self._content_length = int(value)
         self._outheaders.append((name, unicode(value).encode('utf-8')))
 
     def end_headers(self):
-        """Must be called after all headers have been sent and before the actual
-        content is written.
+        """Must be called after all headers have been sent and before the
+        actual content is written.
         """
         self._send_cookie_headers()
         self._write = self._start_response(self._status, self._outheaders)
@@ -250,20 +482,22 @@ class Request(object):
             extra = m.hexdigest()
         etag = 'W/"%s/%s/%s"' % (self.authname, http_date(datetime), extra)
         inm = self.get_header('If-None-Match')
-        if (not inm or inm != etag):
+        if not inm or inm != etag:
             self.send_header('ETag', etag)
         else:
             self.send_response(304)
+            self.send_header('Content-Length', 0)
             self.end_headers()
             raise RequestDone
 
     def redirect(self, url, permanent=False):
-        """Send a redirect to the client, forwarding to the specified URL. The
-        `url` may be relative or absolute, relative URLs will be translated
+        """Send a redirect to the client, forwarding to the specified URL.
+
+        The `url` may be relative or absolute, relative URLs will be translated
         appropriately.
         """
-        if self.session:
-            self.session.save() # has to be done before the redirect is sent
+        for listener in self.redirect_listeners:
+            listener(self, url, permanent)
 
         if permanent:
             status = 301 # 'Moved Permanently'
@@ -273,46 +507,31 @@ class Request(object):
             status = 302 # 'Found' -- normal temporary redirect
 
         self.send_response(status)
-        if not url.startswith('http://') and not url.startswith('https://'):
+        if not url.startswith(('http://', 'https://')):
             # Make sure the URL is absolute
             scheme, host = urlparse.urlparse(self.base_url)[:2]
             url = urlparse.urlunparse((scheme, host, url, None, None, None))
 
+        # Workaround #10382, IE6-IE9 bug when post and redirect with hash
+        if status == 303 and '#' in url:
+            match = re.search(' MSIE ([0-9]+)',
+                              self.environ.get('HTTP_USER_AGENT', ''))
+            if match and int(match.group(1)) < 10:
+                url = url.replace('#', '#__msie303:')
+
         self.send_header('Location', url)
         self.send_header('Content-Type', 'text/plain')
+        self.send_header('Content-Length', 0)
         self.send_header('Pragma', 'no-cache')
-        self.send_header('Cache-control', 'no-cache')
+        self.send_header('Cache-Control', 'no-cache')
         self.send_header('Expires', 'Fri, 01 Jan 1999 00:00:00 GMT')
         self.end_headers()
-
-        if self.method != 'HEAD':
-            self.write('Redirecting...')
         raise RequestDone
 
-    def display(self, template, content_type='text/html', status=200):
-        """Render the response using the ClearSilver template given by the
-        `template` parameter, which can be either the name of the template file,
-        or an already parsed `neo_cs.CS` object.
-        """
-        assert self.hdf, 'HDF dataset not available. Check your clearsilver installation'
-        if self.args.has_key('hdfdump'):
-            # FIXME: the administrator should probably be able to disable HDF
-            #        dumps
-            self.perm.require('TRAC_ADMIN')
-            content_type = 'text/plain'
-            data = str(self.hdf)
-        else:
-            try:
-                form_token = self.form_token
-            except AttributeError:
-                form_token = None
-            data = self.hdf.render(template, form_token)
-
-        self.send(data, content_type, status)
-
     def send(self, content, content_type='text/html', status=200):
         self.send_response(status)
-        self.send_header('Cache-control', 'must-revalidate')
+        self.send_header('Cache-Control', 'must-revalidate')
+        self.send_header('Expires', 'Fri, 01 Jan 1999 00:00:00 GMT')
         self.send_header('Content-Type', content_type + ';charset=utf-8')
         self.send_header('Content-Length', len(content))
         self.end_headers()
@@ -324,31 +543,32 @@ class Request(object):
     def send_error(self, exc_info, template='error.html',
                    content_type='text/html', status=500, env=None, data={}):
         try:
-            if template.endswith('.cs') and self.hdf: # FIXME: remove this
-                if self.args.has_key('hdfdump'):
-                    self.perm.require('TRAC_ADMIN')
-                    content_type = 'text/plain'
-                    data = str(self.hdf)
-                else:
-                    data = self.hdf.render(template)
-
             if template.endswith('.html'):
                 if env:
                     from trac.web.chrome import Chrome
-                    data = Chrome(env).render_template(self, template, data,
-                                                       'text/html')
+                    try:
+                        data = Chrome(env).render_template(self, template,
+                                                           data, 'text/html')
+                    except Exception:
+                        # second chance rendering, in "safe" mode
+                        data['trac_error_rendering'] = True
+                        data = Chrome(env).render_template(self, template,
+                                                           data, 'text/html')
                 else:
                     content_type = 'text/plain'
                     data = '%s\n\n%s: %s' % (data.get('title'),
                                              data.get('type'),
                                              data.get('message'))
-        except: # failed to render
+        except Exception: # failed to render
             data = get_last_traceback()
             content_type = 'text/plain'
 
+        if isinstance(data, unicode):
+            data = data.encode('utf-8')
+
         self.send_response(status)
         self._outheaders = []
-        self.send_header('Cache-control', 'must-revalidate')
+        self.send_header('Cache-Control', 'must-revalidate')
         self.send_header('Expires', 'Fri, 01 Jan 1999 00:00:00 GMT')
         self.send_header('Content-Type', content_type + ';charset=utf-8')
         self.send_header('Content-Length', len(data))
@@ -361,9 +581,16 @@ class Request(object):
             self.write(data)
         raise RequestDone
 
+    def send_no_content(self):
+        self.send_response(204)
+        self.send_header('Content-Length', 0)
+        self.send_header('Content-Type', 'text/plain')
+        self.end_headers()
+        raise RequestDone
+
     def send_file(self, path, mimetype=None):
         """Send a local file to the browser.
-        
+
         This method includes the "Last-Modified", "Content-Type" and
         "Content-Length" headers in the response, corresponding to the file
         attributes. It also checks the last modification time of the local file
@@ -371,13 +598,14 @@ class Request(object):
         "304 Not Modified" response if it matches.
         """
         if not os.path.isfile(path):
-            raise HTTPNotFound("File %s not found" % path)
+            raise HTTPNotFound(_("File %(path)s not found", path=path))
 
         stat = os.stat(path)
         mtime = datetime.fromtimestamp(stat.st_mtime, localtz)
         last_modified = http_date(mtime)
         if last_modified == self.get_header('If-Modified-Since'):
             self.send_response(304)
+            self.send_header('Content-Length', 0)
             self.end_headers()
             raise RequestDone
 
@@ -389,9 +617,12 @@ class Request(object):
         self.send_header('Content-Type', mimetype)
         self.send_header('Content-Length', stat.st_size)
         self.send_header('Last-Modified', last_modified)
+        use_xsendfile = getattr(self, 'use_xsendfile', False)
+        if use_xsendfile:
+            self.send_header('X-Sendfile', os.path.abspath(path))
         self.end_headers()
 
-        if self.method != 'HEAD':
+        if not use_xsendfile and self.method != 'HEAD':
             fileobj = file(path, 'rb')
             file_wrapper = self.environ.get('wsgi.file_wrapper', _FileWrapper)
             self._response = file_wrapper(fileobj, 4096)
@@ -412,23 +643,38 @@ class Request(object):
     def write(self, data):
         """Write the given data to the response body.
 
-        `data` can be either a `str` or an `unicode` string.
-        If it's the latter, the unicode string will be encoded
-        using the charset specified in the ''Content-Type'' header
-        or 'utf-8' otherwise.
+        *data* **must** be a `str` string, encoded with the charset
+        which has been specified in the ``'Content-Type'`` header
+        or UTF-8 otherwise.
+
+        Note that when the ``'Content-Length'`` header is specified,
+        its value either corresponds to the length of *data*, or, if
+        there are multiple calls to `write`, to the cumulative length
+        of the *data* arguments.
         """
         if not self._write:
             self.end_headers()
         if isinstance(data, unicode):
-            data = data.encode(self._outcharset or 'utf-8')
-        self._write(data)
+            raise ValueError("Can't send unicode content")
+        try:
+            self._write(data)
+        except (IOError, socket.error), e:
+            if e.args[0] in (errno.EPIPE, errno.ECONNRESET, 10053, 10054):
+                raise RequestDone
+            # Note that mod_wsgi raises an IOError with only a message
+            # if the client disconnects
+            if 'mod_wsgi.version' in self.environ and \
+               e.args[0] in ('failed to write data',
+                             'client connection closed'):
+                raise RequestDone
+            raise
 
     # Internal methods
 
-    def _parse_args(self):
-        """Parse the supplied request parameters into a dictionary."""
-        args = _RequestArgs()
-
+    def _parse_arg_list(self):
+        """Parse the supplied request parameters into a list of
+        `(name, value)` tuples.
+        """
         fp = self.environ['wsgi.input']
 
         # Avoid letting cgi.FieldStorage consume the input stream when the
@@ -440,23 +686,21 @@ class Request(object):
                          'multipart/form-data'):
             fp = StringIO('')
 
-        fs = cgi.FieldStorage(fp, environ=self.environ, keep_blank_values=True)
-        if fs.list:
-            for name in fs.keys():
-                values = fs[name]
-                if not isinstance(values, list):
-                    values = [values]
-                for value in values:
-                    if not value.filename:
-                        value = unicode(value.value, 'utf-8')
-                    if name in args:
-                        if isinstance(args[name], list):
-                            args[name].append(value)
-                        else:
-                            args[name] = [args[name], value]
-                    else:
-                        args[name] = value
+        # Python 2.6 introduced a backwards incompatible change for
+        # FieldStorage where QUERY_STRING is no longer ignored for POST
+        # requests. We'll keep the pre 2.6 behaviour for now...
+        if self.method == 'POST':
+            qs_on_post = self.environ.pop('QUERY_STRING', '')
+        fs = _FieldStorage(fp, environ=self.environ, keep_blank_values=True)
+        if self.method == 'POST':
+            self.environ['QUERY_STRING'] = qs_on_post
 
+        args = []
+        for value in fs.list or ():
+            name = value.name
+            if not value.filename:
+                value = unicode(value.value, 'utf-8')
+            args.append((name, value))
         return args
 
     def _parse_cookies(self):
@@ -476,6 +720,24 @@ class Request(object):
             headers.append(('content-type', self.environ['CONTENT_TYPE']))
         return headers
 
+    def _parse_languages(self):
+        """The list of languages preferred by the remote user, taken from the
+        ``Accept-Language`` header.
+        """
+        header = self.get_header('Accept-Language') or 'en-us'
+        langs = []
+        for i, lang in enumerate(header.split(',')):
+            code, params = cgi.parse_header(lang)
+            q = 1
+            if 'q' in params:
+                try:
+                    q = float(params['q'])
+                except ValueError:
+                    q = 0
+            langs.append((-q, i, code))
+        langs.sort()
+        return [code for q, i, code in langs]
+
     def _reconstruct_url(self):
         """Reconstruct the absolute base URL of the application."""
         host = self.get_header('Host')
@@ -483,7 +745,8 @@ class Request(object):
             # Missing host header, so reconstruct the host from the
             # server name and port
             default_port = {'http': 80, 'https': 443}
-            if self.server_port and self.server_port != default_port[self.scheme]:
+            if self.server_port and self.server_port != \
+                    default_port[self.scheme]:
                 host = '%s:%d' % (self.server_name, self.server_port)
             else:
                 host = self.server_name
@@ -499,98 +762,8 @@ class Request(object):
                            .replace(',', '%3C')
             self.outcookie[name]['path'] = path
 
-        cookies = self.outcookie.output(header='')
+        cookies = to_unicode(self.outcookie.output(header='')).encode('utf-8')
         for cookie in cookies.splitlines():
             self._outheaders.append(('Set-Cookie', cookie.strip()))
 
-
-class IAuthenticator(Interface):
-    """Extension point interface for components that can provide the name
-    of the remote user."""
-
-    def authenticate(req):
-        """Return the name of the remote user, or `None` if the identity of the
-        user is unknown."""
-
-
-class IRequestHandler(Interface):
-    """Extension point interface for request handlers."""
-
-    def match_request(req):
-        """Return whether the handler wants to process the given request."""
-
-    def process_request(req):
-        """Process the request. For ClearSilver, return a (template_name,
-        content_type) tuple, where `template` is the ClearSilver template to use
-        (either a `neo_cs.CS` object, or the file name of the template), and
-        `content_type` is the MIME type of the content. For Genshi, return a
-        (template_name, data, content_type) tuple, where `data` is a dictionary
-        of substitutions for the template.
-
-        For both templating systems, "text/html" is assumed if `content_type` is
-        `None`.
-
-        Note that if template processing should not occur, this method can
-        simply send the response itself and not return anything.
-        """
-
-
-class IRequestFilter(Interface):
-    """Extension point interface for components that want to filter HTTP
-    requests, before and/or after they are processed by the main handler."""
-
-    def pre_process_request(req, handler):
-        """Called after initial handler selection, and can be used to change
-        the selected handler or redirect request.
-        
-        Always returns the request handler, even if unchanged.
-        """
-
-    # for ClearSilver templates
-    def post_process_request(req, template, content_type):
-        """Do any post-processing the request might need; typically adding
-        values to req.hdf, or changing template or mime type.
-        
-        Always returns a tuple of (template, content_type), even if
-        unchanged.
-
-        Note that `template`, `content_type` will be `None` if:
-         - called when processing an error page
-         - the default request handler did not return any result
-
-        (for 0.10 compatibility; only used together with ClearSilver templates)
-        """
-
-    # for Genshi templates
-    def post_process_request(req, template, data, content_type):
-        """Do any post-processing the request might need; typically adding
-        values to the template `data` dictionary, or changing template or
-        mime type.
-        
-        `data` may be update in place.
-
-        Always returns a tuple of (template, data, content_type), even if
-        unchanged.
-
-        Note that `template`, `data`, `content_type` will be `None` if:
-         - called when processing an error page
-         - the default request handler did not return any result
-
-        (Since 0.11)
-        """
-
-
-class ITemplateStreamFilter(Interface):
-    """Filter a Genshi event stream prior to rendering."""
-
-    def filter_stream(req, method, filename, stream, data):
-        """Return a filtered Genshi event stream, or the original unfiltered
-        stream if no match.
-
-        `req` is the current request object, `method` is the Genshi render
-        method (xml, xhtml or text), `filename` is the filename of the template
-        to be rendered, `stream` is the event stream and `data` is the data for
-        the current template.
-
-        See the Genshi documentation for more information.
-        """
+__no_apidoc__ = _HTTPException_subclass_names
